""" Beam Arrival Monitor related sub-routines

    Copyright (2021) SCS Team.

    (contributions preferrably comply with pep8 code structure
    guidelines.)
"""

import logging

import numpy as np
import xarray as xr

from ..misc.bunch_pattern_external import is_pulse_at
from ..mnemonics_machinery import (mnemonics_to_process,
                                   mnemonics_for_run)
from ..constants import mnemonics as _mnemonics

__all__ = [
    'get_bam',
    'get_bam_params',
]

log = logging.getLogger(__name__)


def get_bam(run, mnemonics=None, merge_with=None, bunchPattern='sase3'):
    """
    Load beam arrival monitor (BAM) data and align their pulse ID
    according to the bunch pattern. Sources can be loaded on the fly
    via the mnemonics argument, or processed from an existing data set
    (merge_with).

    Parameters
    ----------
    run: extra_data.DataCollection
        DataCollection containing the bam data.
    mnemonics: str or list of str
        mnemonics for BAM, e.g. "BAM1932M" or ["BAM414", "BAM1932M"].
        If None, defaults to "BAM1932M" in case no merge_with dataset
        is provided.
    merge_with: xarray Dataset
        If provided, the resulting Dataset will be merged with this
        one. The BAM variables of merge_with (if any) will also be
        selected, aligned and merged.
    bunchPattern: str
        'sase1' or 'sase3' or 'scs_ppl', bunch pattern
        used to extract peaks. The pulse ID dimension will be named
        'sa1_pId', 'sa3_pId' or 'ol_pId', respectively.
    Returns
    -------
    xarray Dataset with pulse-resolved BAM variables aligned,
        merged with Dataset *merge_with* if provided.

    Example
    -------
    >>> import toolbox_scs as tb
    >>> import toolbox_scs.detectors as tbdet
    >>> run, _ = tb.load(2711, 303)
    >>> bam = tbdet.get_bam(run)

    """
    # get the list of mnemonics to process
    mnemonics = mnemonics_to_process(mnemonics, merge_with, 'BAM')

    if len(mnemonics) == 0:
        log.info('No array with unaligned BAM data to extract. Skipping.')
        return merge_with
    else:
        log.info(f'Extracting BAM data from {mnemonics}.')

    # Prepare the dataset of non-BAM data to merge with
    if bool(merge_with):
        mw_ds = merge_with.drop(mnemonics, errors='ignore')
    else:
        mw_ds = xr.Dataset()

    run_mnemonics = mnemonics_for_run(run)
    # extract the XFEL pulses: slice(0,5400,2)
    roi = np.s_[:5400:2]
    ds = xr.Dataset()
    for m in mnemonics:
        if bool(merge_with) and m in merge_with:
            val = merge_with[m].isel({run_mnemonics[m]['dim'][0]: roi})
            log.debug(f'Using {m} from merge_with dataset.')
        else:
            val = run.get_array(*run_mnemonics[m].values(), roi=roi, name=m)
            log.debug(f'Loading {m} from DataCollection.')
        val[run_mnemonics[m]['dim'][0]] = np.arange(2700)
        ds = ds.merge(val, join='inner')

    # check if bunch pattern table exists
    if bool(merge_with) and 'bunchPatternTable' in merge_with:
        bpt = merge_with['bunchPatternTable']
        log.debug('Using bpt from merge_with dataset.')
    elif 'bunchPatternTable' in run_mnemonics:
        bpt = run.get_array(*run_mnemonics['bunchPatternTable'].values())
        log.debug('Loaded bpt from extra_data run.')
    else:
        bpt = None

    # align the pulse Id
    if bpt is not None and len(ds.variables) > 0:
        dim_names = {'sase3': 'sa3_pId', 'sase1': 'sa1_pId',
                     'scs_ppl': 'ol_pId'}
        mask = is_pulse_at(bpt, bunchPattern)
        mask = mask.rename({'pulse_slot': dim_names[bunchPattern]})
        ds = ds.rename({run_mnemonics['BAM1932M']['dim'][0]:
                        dim_names[bunchPattern]})
        ds = ds.where(mask, drop=True)

    # merge with non-BAM dataset
    ds = mw_ds.merge(ds, join='inner')

    return ds


def get_bam_params(run, mnemo_or_source='BAM1932S'):
    """
    Extract the run values of bamStatus[1-3] and bamError.

    Parameters
    ----------
    run: extra_data.DataCollection
        DataCollection containing the bam data.
    mnemo_or_source: str
        mnemonic of the BAM, e.g. 'BAM414', or source name,
        e.g. 'SCS_ILH_LAS/DOOCS/BAM_414_B2.

    Returns
    -------
    params: dict
        dictionnary containing the extracted parameters.

    Note
    ----
    The extracted parameters are run values, they do not reflect any
    possible change during the run.
    """

    if mnemo_or_source in _mnemonics:
        run_mnemonics = mnemonics_for_run(run)
        source = run_mnemonics[mnemo_or_source]['source'].split(':')[0]
    else:
        source = mnemo_or_source
    res = {}
    res['status1'] = run.get_run_value(source, 'bamStatus1.value')
    res['status2'] = run.get_run_value(source, 'bamStatus2.value')
    res['status3'] = run.get_run_value(source, 'bamStatus3.value')
    res['error'] = run.get_run_value(source, 'bamError.value')

    return res
